"""OpenSSL hash algorithms.

This module provides a wrapper for OpenSSL hash functions to have a unique
interface for all types of algorithms.

Each hash algorithm object has the methods `update` and the properties
`checksum` & `hash_algorithm`

"""


import hashlib


class OpenSSLHashAlgorithm(object):
    """OpenSSL hash algorithm."""

    # Currently supported algorithms
    SUPPORTED_ALGORITHMS = sorted(["MD5",
                                   "SHA-1",
                                   "SHA-224",
                                   "SHA-256",
                                   "SHA-384",
                                   "SHA-512",
                                   "SHA3-224",
                                   "SHA3-256",
                                   "SHA3-384",
                                   "SHA3-512"])


    def __init__(self, hash_algorithm):
        """Initialize a OpenSSLHashAlgorithm.

        DIF algorithm naming convention and hashlib algorithm names are
        supported.

        Parameters
        ----------
        hash_algorithm : str
            one of `OpenSSLHashAlgorithm.SUPPORTED_ALGORITHMS`

        """

        self.hash_algorithm = hash_algorithm.upper().replace("_", "-")
        hashlib_name = self.hash_algorithm

        # DIF names deviate from python's hashlib names
        deviating_names = [("MD5", "md5"),
                           ("SHA-1", "sha1"),
                           ("SHA-224", "sha224"),
                           ("SHA-256", "sha256"),
                           ("SHA-384", "sha384"),
                           ("SHA-512", "sha512"),
                           ("SHA3-224", "sha3_224"),
                           ("SHA3-256", "sha3_256"),
                           ("SHA3-384", "sha3_384"),
                           ("SHA3-512", "sha3_512")]
        for dif_name, lib_name in deviating_names:
            if self.hash_algorithm == dif_name or\
                    self.hash_algorithm == lib_name.upper():
                self.hash_algorithm = dif_name
                hashlib_name = lib_name
                break

        if self.hash_algorithm not in self.SUPPORTED_ALGORITHMS:
            raise ValueError("{0} is not a supported hash algorithm.".format(
                self.hash_algorithm))

        self._hasher = hashlib.new(hashlib_name)


    def update(self, data):
        """Update the hash.

        Parameters
        ----------
        data : bytes
            the data to update the hash with

        """

        self._hasher.update(data)

    @property
    def checksum(self):
        return self._hasher.hexdigest()

    @property
    def digest_size(self):
        return self._hasher.digest_size



"""Data Intergrity Fingerprint (DIF) programming library.

Import as `dataintegrityfingerprint`.

"""


import os
import codecs
import multiprocessing

from .openssl_hash_algorithm import OpenSSLHashAlgorithm
from .zlib_hash_algorithm import ZlibHashAlgorithm


class DataIntegrityFingerprint:
    """A class representing a DataIntegrityFingerprint (DIF).

    Example
    -------
    dif = DataIntegrityFingerprint("~/Downloads")
    print(dif)
    print(dif.checksums)

    """

    CRYPTOGRAPHIC_ALGORITHMS = OpenSSLHashAlgorithm.SUPPORTED_ALGORITHMS
    NON_CRYPTOGRAPHIC_ALGORITHMS = ZlibHashAlgorithm.SUPPORTED_ALGORITHMS
    CHECKSUM_FILENAME_SEPARATOR = "  "

    def __init__(self, data, from_checksums_file=False,
                 hash_algorithm="SHA-256", multiprocessing=True,
                 allow_non_cryptographic_algorithms=False):
        """Create a DataIntegrityFingerprint object.

        Parameters
        ----------
        data : str
            the path to the data
        from_checksums_file : bool
            data argument is a checksums file
        hash_algorithm : str
            the hash algorithm (optional, default: sha256)
        multiprocessing : bool
            using multi CPU cores (optional, default: True)
            speeds up creating of checksums for large data files
        allow_non_cryptographic_algorithms : bool
            set True only, if you need non cryptographic algorithms (see
            notes!)

        Note
        ----
        We do not suggest to use non-cryptographic algorithms.
        Non-cryptographic algorithms are, while much faster, not secure (e.g.
        can be tempered with). Only use these algorithms to check for technical
        file damage and in cases security is not of critical concern.

        """

        if not from_checksums_file:
            assert os.path.isdir(data)

        h = new_hash_instance(hash_algorithm,
                              allow_non_cryptographic_algorithms)
        self._hash_algorithm = h.hash_algorithm
        self._data = os.path.abspath(data)
        self._file_count = None
        self._hash_list = []
        self._multiprocessing = multiprocessing
        self._allow_non_cryptographic_algorithms = \
            allow_non_cryptographic_algorithms

    def __str__(self):
        return str(self.dif)

    def get_files(self):
        """Get all files to hash.

        Returns
        -------
        files : list
            the list of files to hash

        """

        rtn = []
        if os.path.isdir(self._data):
            for dir_, _, files in os.walk(self._data, followlinks=True):
                for filename in files:
                    rtn.append(os.path.join(dir_, filename))
        self._file_count = len(rtn)
        return rtn

    @ property
    def hash_algorithm(self):
        return self._hash_algorithm

    @property
    def data(self):
        return self._data

    @property
    def file_hash_list(self):
        if len(self._hash_list) < 1:
            self.generate()
        return self._hash_list

    @property
    def checksums(self):
        rtn = ""
        for h, fl in sorted(self.file_hash_list, key=lambda x: x[1]):
            rtn += u"{0}{1}{2}\n".format(h, self.CHECKSUM_FILENAME_SEPARATOR,
                                         fl)
        return rtn

    @property
    def dif(self):
        if len(self.file_hash_list) < 1:
            return None

        hasher = new_hash_instance(self._hash_algorithm,
                                   self.allow_non_cryptographic_algorithms)
        concat = "".join(["".join(x) for x in self.file_hash_list])
        hasher.update(concat.encode("utf-8"))
        return hasher.checksum

    @property
    def file_count(self):
        if self._file_count is None:
            self.get_files()
        return self._file_count

    @property
    def multiprocessing(self):
        return self._multiprocessing

    @property
    def allow_non_cryptographic_algorithms(self):
        return self._allow_non_cryptographic_algorithms

    def generate(self, progress=None):
        """Generate hash list to get Data Integrity Fingerprint.

        Parameters
        ----------
        progress: function, optional
            a callback function for a progress reporting that takes the
            following parameters:
                count  -- the current count
                total  -- the total count
                status -- a string describing the status

        """

        hash_list = []

        if os.path.isfile(self._data):
            # from  checksum file
            with codecs.open(self._data, encoding="utf-8") as f:
                for line in f:
                    h, fl = line.split(self.CHECKSUM_FILENAME_SEPARATOR,
                                       maxsplit=1)
                    hash_list.append((h, fl.strip()))
        else:
            files = self.get_files()
            func_args = zip(files, [self._hash_algorithm] * len(files))
            if self.multiprocessing:
                pool = multiprocessing.Pool()
                imap = pool.imap_unordered
            else:
                imap = map

            for counter, rtn in enumerate(imap(_hash_file_content, func_args)):
                if progress is not None:
                    progress(counter + 1, len(files),
                             "{0}/{1}".format(counter + 1, len(files)))
                fl = os.path.relpath(rtn[1], self.data).replace(os.path.sep,"/")
                hash_list.append((rtn[0], fl))

            if self.multiprocessing:
                pool.close()

        self._hash_list = sorted(hash_list, key=lambda x: x[0] + x[1])

    def save_checksums(self, filename=None):
        """Save the checksums to a file.

        Parameters
        ----------
        filename : str, optional
            the name of the file to save checksums to

        Returns
        -------
        success : bool
            whether saving was successful

        """

        if self.dif is not None:
            if filename is None:
                filename = os.path.split(self.data)[-1] + ".{0}".format(
                    self._hash_algorithm)

            with codecs.open(filename, 'w', "utf-8") as f:
                f.write(self.checksums)

            return True

    def diff_checksums(self, filename):
        """Calculate differences of checksums to checksums file.

        Parameters
        ----------
        filename : str
            the name of the checksums file

        Returns
        -------
        diff : str
            the difference of checksums to the checksums file
            (minus means checksums is missing something from checksums file,
            plus means checksums has something in addition to checksums file)

        """

        checksums = self.checksums.split("\n")
        other = DataIntegrityFingerprint(
            filename, from_checksums_file=True,
            hash_algorithm=self._hash_algorithm,
            allow_non_cryptographic_algorithms=\
                self.allow_non_cryptographic_algorithms)
        checksums_other = other.checksums.split("\n")
        sub = ["- " + x for x in list(set(checksums_other) - set(checksums))]
        add = ["+ " + x for x in list(set(checksums) - set(checksums_other))]

        return "\n".join(["\n".join(sub), "\n".join(add)]).strip()


def new_hash_instance(hash_algorithm,
                      support_non_cryptographic_algorithms=False):
    """Return a new instance of a hash object (similar to hashlib.new()).

    Each HashAlgorithm object has the methods `update` and the
    properties `hash_algorithm` (according the DIF naming convention),
    `checksum`

    Parameters
    ----------
    hash_algorithm : str
        one of `DataIntegrityFingerprint.CRYPTOGRAPHIC_ALGORITHMS`
        (or `DataIntegrityFingerprint.NON_CRYPTOGRAPHIC_ALGORITHMS`)
    support_non_cryptographic_algorithms : bool
        if True, also allow hash algorithms from
        `DataIntegrityFingerprint.NON_CRYPTOGRAPHIC_ALGORITHMS`

    Returns
    -------
    hasher : `ZlibHashAlgorithm` or `OpenSSLHashAlgorithm` object

    """

    if support_non_cryptographic_algorithms:
        try:
            return ZlibHashAlgorithm(hash_algorithm)
        except Exception:
            pass

    try:
        return OpenSSLHashAlgorithm(hash_algorithm)
    except Exception:
        pass


def _hash_file_content(args):
    # args = (filename, hash_algorithm)
    # helper function for multi threading of file hashing
    hasher = new_hash_instance(hash_algorithm=args[1],
                               support_non_cryptographic_algorithms=True)
    with open(args[0], 'rb') as f:
        for block in iter(lambda: f.read(64 * 1024), b''):
            hasher.update(block)

    return hasher.checksum, args[0]



"""Data Integrity Fingerprint (DIF).

Reference Python implementation.

For specification see https://github.com/expyriment/DIF.

"""

__author__ = 'Oliver Lindemann <oliver@expyriment.org>, ' \
             'Florian Krause <florian@expyriment.org>'

__version__ = '0.7.3'


from .dif import DataIntegrityFingerprint



""" Zlib hash algorithms.

This module provides a wrapper for zlib hash functions to have a unique
interface for all types of algorithms.

Each hash algorithm object has the methods `update` and the properties
`checksum` & `hash_algorithm`

"""

import zlib


class ZlibHashAlgorithm(object):
    """Zlib hash algorithm."""

    SUPPORTED_ALGORITHMS = sorted(["CRC-32",
                                   "ADLER-32"])

    def __init__(self, hash_algorithm):
        """Initialize a ZlibHashAlgorithm.

        DIF algorithm naming convention and hashlib algorithm names are
        supported.

        Parameters
        ----------
        hash_algorithm : str
            one of `ZlibHashAlgorithm.SUPPORTED_ALGORITHMS`

        """

        hash_algorithm = hash_algorithm.upper().replace("_", "-")
        if hash_algorithm == "CRC-32":
            self._current = 0
            self._hasher = zlib.crc32
        elif hash_algorithm == "ADLER-32":
            self._current = None
            self._hasher = zlib.adler32
        else:
            raise ValueError("{0} is not a supported hash algorithm.".format(
                hash_algorithm))
        self.hash_algorithm = hash_algorithm

    def update(self, data):
        """Update the hash.

        Parameters
        ----------
        data : bytes
            the data to update the hash with

        """

        try:
            self._current = self._hasher(data, self._current)
        except:
            self._current = self._hasher(data)

    @property
    def checksum(self):
        return hex(self._current)[2:]



"""Data Intergrity Fingerprint (DIF) command line interface.

Invoke with `python3 -m dataintegrityfingerprint` or
`dataintegrityfingerprint`.

"""


import os
import sys
import argparse

from . import DataIntegrityFingerprint
from . import __version__


def cli():

    def progress(count, total, status=''):
        bar_len = 50
        filled_len = int(round(bar_len * count / float(total)))
        percents = round(100.0 * count / float(total), 1)
        bar = '=' * filled_len + ' ' * (bar_len - filled_len)
        sys.stdout.write('{:5.1f}% [{}] {}\r'.format(percents, bar, status))
        sys.stdout.flush()

    parser = argparse.ArgumentParser(
        description="""Data Integrity Fingerprint (DIF)
Reference Python implementation v{0}""".format(__version__),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""Authors:
Oliver Lindemann <oliver@expyriment.org>
Florian Krause <florian@expyriment.org""")

    parser.add_argument("PATH", nargs='?', default=None,
                        help="the path to the data directory")
    parser.add_argument("-f", "--from-checksums-file",
                        dest="fromchecksumsfile", action="store_true",
                        help="Calculate dif from checksums file. " +
                             "PATH is a checksums file",
                        default=False)
    parser.add_argument("-a", "--algorithm", metavar="ALGORITHM",
                        type=str,
                        help="the hash algorithm to be used (default=SHA-256)",
                        default="sha256")
    parser.add_argument("-C", "--checksums", dest="checksums",
                        action="store_true",
                        help="print checksums only",
                        default=False)
    parser.add_argument("-D", "--dif-only", dest="difonly",
                        action="store_true",
                        help="print dif only",
                        default=False)
    parser.add_argument("-G", "--gui", dest="gui",
                        action="store_true",
                        help="open graphical user interface",
                        default=False)
    parser.add_argument("-L", "--list-available-algorithms", dest="listalgos",
                        action="store_true",
                        help="print available algorithms",
                        default=False)
    parser.add_argument("-s", "--save-checksums-file",
                        dest="savechecksumsfile", action="store_true",
                        help="save checksums to file",
                        default=False)
    parser.add_argument("-d", "--diff-checksums-file", metavar="CHECKSUMSFILE",
                        type=str,
                        help="Calculate differences of checksums to " +
                             "CHECKSUMSFILE")
    parser.add_argument("-n", "--no-multi-processing", dest="nomultiprocess",
                        action="store_true",
                        help="switch of multi processing",
                        default="")
    parser.add_argument("-p", "--progress", dest="progressbar",
                        action="store_true",
                        help="show progressbar",
                        default=False)
    parser.add_argument("--non-cryptographic",
                        dest="noncrypto",
                        action="store_true",
                        help="allow non cryptographic algorithms " +
                             "(Not suggested, please read documentation " +
                             "carefully!)",
                        default=False)

    args = vars(parser.parse_args())

    if args['listalgos']:
        print("Crypotographic algorithms")
        print("- " + "\n- ".join(
            DataIntegrityFingerprint.CRYPTOGRAPHIC_ALGORITHMS))
        if args['noncrypto']:
            print("Non-crypotographic algorithms")
            print("- " + "\n- ".join(
                DataIntegrityFingerprint.NON_CRYPTOGRAPHIC_ALGORITHMS))
        sys.exit()

    if args['gui']:
        from .gui import start_gui
        start_gui(data_dir=args["PATH"], hash_algorithm=args["algorithm"])
        sys.exit()

    if args["PATH"] is None:
        print("Use -G to launch the GUI or -h for details about command line interface")
        sys.exit()

    dif = DataIntegrityFingerprint(
        data=args["PATH"],
        from_checksums_file=args['fromchecksumsfile'],
        hash_algorithm=args["algorithm"],
        multiprocessing=not(args['nomultiprocess']),
        allow_non_cryptographic_algorithms=args['noncrypto'])

    if not args['fromchecksumsfile'] and args['progressbar']:
        dif.generate(progress=progress)
        print("")

    # Output
    if args['savechecksumsfile']:
        extension = "".join(
            x for x in dif._hash_algorithm.lower() if x.isalnum())
        outfile = os.path.split( dif.data)[-1] + ".{0}".format(extension)
        answer = "y"
        if os.path.exists(outfile):
            answer = input(
                "'{0}' already exists! Overwrite? [y/N]: ".format(outfile))
        if answer == "y":
            dif.save_checksums()
            print("Checksums have been written to '{0}'.".format(outfile))
        else:
            print("Checksums have NOT been written.")

    elif args['diff_checksums_file']:
        diff = dif.diff_checksums(args['diff_checksums_file'])
        if diff != "":
            print(diff)

    elif args['difonly']:
        print(dif)
    elif args['checksums']:
        print(dif.checksums.strip())
    else:
        print("Data Integrity Fingerprint (DIF)")
        print("")
        print("Directory: {0}".format(dif.data))
        print("Files: {0}".format(dif.file_count))
        print("Algorithm: {0}".format(dif.hash_algorithm))
        print("")
        print("DIF [{}]: {}".format(dif.hash_algorithm, dif))



"""Data Integrity Fingerprint (DIF) graphical user interface.

Invoke with `python3 -m dataintegrityfingerprint.gui` or
`dataintegrityfingerprint -G`.

"""


import os
import platform
import multiprocessing
from threading import Thread

import tkinter as tk
import tkinter.ttk as ttk
from tkinter import filedialog, messagebox

from . import __version__
from .dif import DataIntegrityFingerprint as DIF
from .dif import new_hash_instance

class App(ttk.Frame):
    """The main GUI application."""

    def __init__(self, master, *args, **kwargs):
        """Initialize the GUI application."""

        ttk.Frame.__init__(self, master, *args, **kwargs)
        self.master = master
        self.master.title("Data Integrity Fingerprint (DIF)")
        self.about_text = """Data Integrity Fingerprint (DIF)
Reference Python implementation v{0}

Authors:
Oliver Lindemann <oliver@expyriment.org>
Florian Krause <florian@expyriment.org>
""".format(__version__)
        self.dif = None
        self.create_widgets()
        self.dir_button.focus()

    def create_widgets(self):
        """Create GUI widgets."""

        # Menu
        self.menubar = tk.Menu(self.master)
        if platform.system() == "Darwin":
            modifier = "Command"
            self.apple_menu = tk.Menu(self.menubar, name="apple")
            self.menubar.add_cascade(menu=self.apple_menu)
            self.apple_menu.add_command(
                label="About Data Integrity Fingerprint (DIF)",
                command=lambda: messagebox.showinfo("About", self.about_text))
        else:
            modifier = "Control"
        self.file_menu = tk.Menu(self.menubar)
        self.menubar.add_cascade(menu=self.file_menu, label="File")
        self.file_menu.add_command(label="Open checksums",
                                   command=lambda: Thread(
                                       target=self.open_checksums,
                                       daemon=True).start(),
                                   accelerator="{0}-O".format(modifier))
        self.master.bind("<{0}-o>".format(modifier),
                         lambda event: Thread(target=self.open_checksums,
                                              daemon=True).start())
        self.file_menu.add_command(label="Save checksums",
                                   command=self.save_checksums,
                                   accelerator="{0}-S".format(modifier))
        self.master.bind("<{0}-s>".format(modifier), self.save_checksums)
        self.file_menu.add_command(label="Quit",
                                   command=self.master.destroy,
                                   accelerator="{0}-Q".format(modifier))
        self.master.bind("<{0}-q>".format(modifier),
                         lambda event: self.master.destroy())
        self.file_menu.entryconfig("Save checksums", state=tk.DISABLED)
        self.edit_menu = tk.Menu(self.menubar)
        self.menubar.add_cascade(menu=self.edit_menu, label="Edit")
        self.edit_menu.add_command(label="Diff checksums",
                                   command=self.diff_checksums,
                                   accelerator="{0}-D".format(modifier))
        self.master.bind("<{0}-d>".format(modifier),
                         lambda event: self.diff_checksums())
        self.edit_menu.entryconfig("Diff checksums", state=tk.DISABLED)
        self.options_menu = tk.Menu(self.menubar)
        self.menubar.add_cascade(menu=self.options_menu, label="Options")
        self.algorithm_menu = tk.Menu(self.menubar)
        self.algorithm_var = tk.StringVar()
        self.algorithm_var.set("SHA-256")
        self.algorithm_var.trace('w', self.set_dif_label)
        for algorithm in DIF.CRYPTOGRAPHIC_ALGORITHMS:
            self.algorithm_menu.add_radiobutton(label=algorithm,
                                                value=algorithm,
                                                variable=self.algorithm_var)
        self.options_menu.add_cascade(menu=self.algorithm_menu,
                                      label="Hash algorithm")
        self.update_menu = tk.Menu(self.menubar)
        self.update_var = tk.IntVar()
        self.update_var.set(1)
        self.update_menu.add_radiobutton(label="On", value=1,
                                         variable=self.update_var)
        self.update_menu.add_radiobutton(label="Off (faster)", value=0,
                                         variable=self.update_var)
        self.options_menu.add_cascade(menu=self.update_menu,
                                      label="Progress updating")
        self.multiprocess_var = tk.IntVar()
        self.multiprocess_var.set(0)
        if multiprocessing.cpu_count() > 1:
            self.multiprocess_var.set(1)
            self.multiprocess_menu = tk.Menu(self.menubar)
            self.multiprocess_menu.add_radiobutton(
                label="On ({0} cores)".format(
                    multiprocessing.cpu_count()),
                value=1,
                variable=self.multiprocess_var)
            self.multiprocess_menu.add_radiobutton(
                label="Off", value=0, variable=self.multiprocess_var)
            self.options_menu.add_cascade(menu=self.multiprocess_menu,
                                          label="Multi-core processing")
        self.help_menu = tk.Menu(self.menubar)
        self.menubar.add_cascade(menu=self.help_menu, label="Help")
        self.help_menu.add_command(
            label="About",
            command=lambda: messagebox.showinfo("About", self.about_text))

        self.master["menu"] = self.menubar

        # Main window
        self.frame1 = ttk.Frame(self.master)
        self.frame1.grid(row=0, column=0, sticky="NSWE")
        self.frame1.grid_columnconfigure(1, weight=1)
        self.dir_label = ttk.Label(self.frame1, text="Data directory:")
        self.dir_label.grid(row=0, column=0)
        self.dir_var = tk.StringVar()
        self.dir_var.set("")
        self.dir_entry = ttk.Entry(self.frame1, textvariable=self.dir_var,
                                   takefocus=0, state="readonly")
        self.dir_entry.grid(row=0, column=1, sticky="WE")
        self.dir_button = ttk.Button(self.frame1, text="Browse...",
                                     command=self.set_data_directory)
        self.dir_button.grid(row=0, column=2)
        self.generate_button = ttk.Button(
            self.frame1, text="Generate DIF",
            command=lambda: Thread(target=self.generate_dif,
                                   daemon=True).start(),
            state=tk.DISABLED)
        self.generate_button.grid(row=0, column=3)

        self.progressbar = ttk.Progressbar(self.master)
        self.progressbar.grid(row=1, column=0, sticky="NSWE")

        self.container = ttk.Frame(self.master, borderwidth=1,
                                   relief=tk.SUNKEN)
        self.checksum_list = tk.Text(self.container, wrap="none",
                                     borderwidth=0, state=tk.DISABLED)
        self.checksum_list.bind("<1>",
                                lambda event: self.checksum_list.focus_set())
        self.vertical_scroll = ttk.Scrollbar(self.container, orient="vertical",
                                             command=self.checksum_list.yview)
        self.horizontal_scroll = ttk.Scrollbar(
            self.container, orient="horizontal",
            command=self.checksum_list.xview)
        self.checksum_list.configure(yscrollcommand=self.vertical_scroll.set,
                                     xscrollcommand=self.horizontal_scroll.set)
        self.checksum_list.grid(row=0, column=0, sticky="NSWE")
        self.vertical_scroll.grid(row=0, column=1, sticky="NS")
        self.horizontal_scroll.grid(row=1, column=0, sticky="EW")
        self.container.grid_rowconfigure(0, weight=1)
        self.container.grid_columnconfigure(0, weight=1)
        self.container.grid(row=2, column=0, sticky="NSWE")

        self.frame2 = ttk.Frame(self.master)
        self.frame2.grid(row=3, column=0, sticky="NSWE")
        self.frame2.grid_columnconfigure(1, weight=1)
        self.dif_label = ttk.Label(self.frame2)
        self.dif_label.grid(row=0, column=0)
        self.set_dif_label()
        self.dif_var = tk.StringVar()
        self.dif_var.set("")
        self.dif_entry = ttk.Entry(self.frame2, textvariable=self.dif_var,
                                   takefocus=0, foreground="black",
                                   state=tk.DISABLED)
        self.dif_entry.grid(row=0, column=1, sticky="NSWE")
        self.copy_button = ttk.Button(self.frame2, text="Copy",
                                      command=self.copy_dif_to_clipboard,
                                      state=tk.DISABLED)
        self.copy_button.grid(row=0, column=2)
        self.copy_button.bind("<Button-2>",
                              lambda e: self.copy_dif_to_clipboard("badge"))

        # Status bar
        self.statusbar = ttk.Label(self.master, text="", border=1,
                                   relief=tk.SUNKEN, anchor=tk.W)
        self.statusbar.grid(row=4, column=0, sticky="WE")

    def set_data_directory(self, data_dir=None, *args):
        """Set the data directory."""

        if data_dir is None:
            data_dir = filedialog.askdirectory()
        if os.path.isdir(data_dir):
            self.dir_var.set(data_dir)
            self.generate_button["state"] = tk.NORMAL
            self.generate_button.focus()
            self.progressbar["value"] = 0
            self.checksum_list["state"] = tk.NORMAL
            self.checksum_list.delete(1.0, tk.END)
            self.checksum_list["state"] = tk.DISABLED
            self.dif_var.set("")
            self.dif = None
            self.statusbar["text"] = "Ready"
            self.unblock_gui(enable_save_checksums=False)

    def set_dif_label(self, *args):
        self.dif_label.config(text="DIF [{0}]:".format(
            self.algorithm_var.get()))

    def block_gui(self):
        """Block GUI from user entry."""

        self.file_menu.entryconfig("Open checksums", state=tk.DISABLED)
        self.file_menu.entryconfig("Save checksums", state=tk.DISABLED)
        self.options_menu.entryconfig("Hash algorithm", state=tk.DISABLED)
        self.options_menu.entryconfig("Progress updating", state=tk.DISABLED)
        self.options_menu.entryconfig("Multi-core processing",
                                      state=tk.DISABLED)
        self.help_menu.entryconfig("About", state=tk.DISABLED)
        self.progressbar.grab_set()

    def unblock_gui(self, enable_save_checksums=True):
        """Unblock GUI from user entry."""

        self.file_menu.entryconfig("Open checksums", state=tk.NORMAL)
        if enable_save_checksums:
            self.file_menu.entryconfig("Save checksums", state=tk.NORMAL)
        self.file_menu.entryconfig("Quit", state=tk.NORMAL)
        self.edit_menu.entryconfig("Diff checksums", state=tk.NORMAL)
        self.options_menu.entryconfig("Hash algorithm", state=tk.NORMAL)
        self.options_menu.entryconfig("Progress updating", state=tk.NORMAL)
        self.options_menu.entryconfig("Multi-core processing",
                                      state=tk.NORMAL)
        self.help_menu.entryconfig("About", state=tk.NORMAL)
        self.progressbar.grab_release()

    def generate_dif(self, *args):
        """Generate DIF from data directory"""

        def _progress(count, total, status=''):
            """Progress callback function"""

            percents = int(100.0 * count / float(total))
            self.progressbar["value"] = percents
            self.statusbar["text"] = \
                "Generating DIF...{0}% ({1})".format(percents, status)

        self.block_gui()
        self.statusbar["text"] = "Generating DIF..."
        if self.multiprocess_var.get() == 1:
            multiprocessing = True
        else:
            multiprocessing = False
        self.dif = DIF(self.dir_entry.get(),
                       hash_algorithm=self.algorithm_var.get(),
                       multiprocessing=multiprocessing)
        if self.update_var.get() == 1:
            progress = _progress
        else:
            progress = None
        self.dif.generate(progress=progress)
        self.progressbar["value"] = 100
        self.checksum_list["state"] = tk.NORMAL
        self.checksum_list.delete(1.0, tk.END)
        self.checksum_list.insert(1.0, self.dif.checksums.strip("\n"))
        self.checksum_list["state"] = tk.DISABLED
        self.dif_var.set(self.dif.dif)
        self.set_dif_label()
        self.copy_button["state"] = tk.NORMAL
        self.copy_button.focus()
        self.statusbar["text"] = "Done"
        self.unblock_gui()

    def open_checksums(self, *args):
        """Open checksums file."""

        allowed_extensions = ""
        for algorithm in DIF.CRYPTOGRAPHIC_ALGORITHMS:
            extension = "".join(x for x in algorithm.lower() if x.isalnum())
            allowed_extensions += "*.{0} ".format(extension)
        filetypes = [("Checksums files", allowed_extensions.strip())]
        filename = filedialog.askopenfilename(filetypes=filetypes)
        if os.path.exists(filename):
            try:
                self.block_gui()
                old_status = self.statusbar["text"]
                old_progress_mode = self.progressbar["mode"]
                old_progress_value = self.progressbar["value"]
                self.progressbar["mode"] = "indeterminate"
                self.progressbar["value"] = 0
                self.progressbar.start()
                self.statusbar["text"] = "Opening checksums..."
                algorithm = os.path.splitext(filename)[-1].strip(".")
                self.dif = DIF(filename, from_checksums_file=True,
                               hash_algorithm=self.algorithm_var.get())
                master_hash = self.dif.dif
                checksums = self.dif.checksums.strip("\n")
                self.dir_var.set("")
                self.generate_button["state"] = tk.DISABLED
                self.progressbar.stop()
                self.progressbar["mode"] = "determinate"
                self.progressbar["value"] = 100
                self.checksum_list["state"] = tk.NORMAL
                self.checksum_list.delete(1.0, tk.END)
                self.checksum_list.insert(1.0, checksums)
                self.checksum_list["state"] = tk.DISABLED
                self.set_dif_label()
                self.dif_var.set(master_hash)
                self.copy_button["state"] = tk.NORMAL
                self.copy_button.focus()
                self.statusbar["text"] = filename
                self.unblock_gui()
            except Exception:
                self.progressbar.stop()
                self.progressbar["mode"] = old_progress_mode
                self.progressbar["value"] = old_progress_value
                self.statusbar["text"] = old_status
                self.unblock_gui()
                messagebox.showerror("Error", "Not a valid checksums file")
                self.unblock_gui()

    def save_checksums(self, *args):
        """Save checksums file."""

        if self.checksum_list.get(1.0, tk.END).strip("\n") != "":
            algorithm = self.algorithm_var.get()
            extension = "".join(x for x in algorithm.lower() if x.isalnum())
            filename = filedialog.asksaveasfilename(
                defaultextension=extension,
                filetypes=[("{0} files".format(algorithm),
                            "*.{0}".format(extension))],
                initialdir=os.path.split(self.dir_entry.get())[0],
                initialfile=os.path.split(self.dir_entry.get())[-1])
            if filename != "":
                self.dif.save_checksums(filename)
            self.unblock_gui()

    def diff_checksums(self, *args):
        """Calculate difference of checksums to checksums file."""

        allowed_extensions = ""
        for algorithm in DIF.CRYPTOGRAPHIC_ALGORITHMS:
            extension = "".join(x for x in algorithm.lower() if x.isalnum())
            allowed_extensions += "*.{0} ".format(extension)
        filetypes = [("Checksums files", allowed_extensions.strip())]
        filename = filedialog.askopenfilename(filetypes=filetypes)
        if os.path.exists(filename):
            diff = self.dif.diff_checksums(filename)
            DiffDialogue(self.master, filename, diff).show()

    def copy_dif_to_clipboard(self, *args):
        self.master.clipboard_clear()
        if "badge" in args and self.dif_var.get() != "":
            label = "DIF [{0}]".format(self.algorithm_var.get().replace("-",
                                                                        "--"))
            self.master.clipboard_append(
                "https://img.shields.io/badge/{0}-{1}-informational".format(
                    label, self.dif_var.get()))
        else:
            self.master.clipboard_append(self.dif_var.get())


class DiffDialogue:
    """The dialogue for the checksums differences."""

    def __init__(self, master, filename, diff):
        """Initialize the dialogue."""

        self.master = master
        top = self.top = tk.Toplevel(master)
        top.title("Differences to {0}".format(filename))

        self.container = ttk.Frame(top, borderwidth=1,
                                   relief=tk.SUNKEN)
        self.checksum_list = tk.Text(self.container, wrap="none",
                                     borderwidth=0)
        self.checksum_list.bind("<1>",
                                lambda event: self.checksum_list.focus_set())
        self.vertical_scroll = ttk.Scrollbar(self.container, orient="vertical",
                                             command=self.checksum_list.yview)
        self.horizontal_scroll = ttk.Scrollbar(
            self.container, orient="horizontal",
            command=self.checksum_list.xview)
        self.checksum_list.configure(yscrollcommand=self.vertical_scroll.set,
                                     xscrollcommand=self.horizontal_scroll.set)
        self.checksum_list.grid(row=0, column=0, sticky="NSWE")
        self.vertical_scroll.grid(row=0, column=1, sticky="NS")
        self.horizontal_scroll.grid(row=1, column=0, sticky="EW")
        self.container.grid_rowconfigure(0, weight=1)
        self.container.grid_columnconfigure(0, weight=1)
        self.container.grid(row=0, column=0, sticky="NSWE")

        self.checksum_list.insert(1.0, diff)
        self.checksum_list["state"] = tk.DISABLED

        top.focus()
        top.bind('<Escape>', lambda x: self.top.destroy())
        top.geometry("1024x400")
        top.grid_columnconfigure(0, weight=1)
        top.grid_rowconfigure(0, weight=1)

    def show(self):
        self.master.wait_window(self.top)

def start_gui(data_dir=None, hash_algorithm=None):
    root = tk.Tk()
    root.bind_class("TButton", "<Return>",
                    lambda event: event.widget.invoke())
    root.option_add('*tearOff', tk.FALSE)
    root.geometry("1024x600")
    root.grid_columnconfigure(0, weight=1)
    root.grid_rowconfigure(2, weight=1)
    app = App(root)

    if data_dir is not None:
        app.set_data_directory(os.path.abspath(data_dir))
    if hash_algorithm is not None:
        h = new_hash_instance(hash_algorithm)
        app.algorithm_var.set(h.hash_algorithm)

    app.mainloop()

if __name__ == "__main__":
    start_gui()



import os
import setuptools


PACKAGE_NAME = "dataintegrityfingerprint"


def get_version():
    """Get version number."""

    with open(os.path.join("src", PACKAGE_NAME, "__init__.py")) as f:
        for line in f:
            if line.startswith("__version__"):
                return line.split("'")[1]
    return "None"

def get_readme():
    with open("README.md", "r", encoding="utf-8") as fh:
        return fh.read()


setuptools.setup(
    name = PACKAGE_NAME,
    version=get_version(),
    author='Oliver Lindemann, Florian Krause',
    author_email='oliver@expyriment.org, florian@expyriment.org',
    description='Data Integrity Fingerprint (DIF) - ' \
    'Reference Python implementation',
    long_description=get_readme(),
    long_description_content_type='text/markdown',
    url="https://github.com/expyriment/dataintegrityfingerprint-python",
    classifiers=[
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering"
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    entry_points = {
        'console_scripts': [f'{PACKAGE_NAME}={PACKAGE_NAME}.cli:cli'],
        'gui_scripts': [f'{PACKAGE_NAME}-gui={PACKAGE_NAME}.gui:start_gui']
    }
)
